@@ -1,4 +1,5 @@
 """Provides an interface for determining how Python objects are serialized and de-serialized."""
+import threading
 
 from amfast import AmFastError
 
@@ -13,22 +14,31 @@ class ClassDef(object):
     attributes
     ===========
      * class_ - class, the class object mapped to this definition
-     * alias - string, the amf alias name of the mapped class
+     * alias - string, the AMF alias name of the mapped class
      * static_attrs - tuple or list, a tuple of static attribute names,
-          all values must be strings or unicode
-     * amf3 - bool, if True, this object will be encoded in AMF3. 
+          all values must be strings or unicode.
+     * amf3 - bool, if True, this object will be encoded in AMF3.
+     * encode_types - dict, keys = attribute names, values = callables.
+         Callables must accept a single parameter
+         (the object being encoded) and return a new object.
+     * decode_types - dict, keys = attribute names, values = callables.
+         Callables must accept a single parameter
+         (the object being decoded) and return a new object.
+
     """
 
     CLASS_DEF = True
 
     def __init__(self, class_, alias=None, static_attrs=None,
-        amf3=None, _built_in=False):
+        amf3=None, encode_types=None, decode_types=None, _built_in=False):
         """arguments
         =============
          * class_ - class, the class being mapped.
          * alias - string, specifies the amf class alias. Default = module.class
          * static_attrs - tuple or list, a tuple of static attribute strings. Default = empty tuple
          * amf3 - bool, if True, this object will be encoded in AMF3. Default = True
+         * encode_types - dict, keys = attribute names, values = callables. Default = empty dict
+         * decode_types - dict, keys = attribute names, values = callables. Default = empty dict
         """
         self.class_ = class_
         self._built_in = _built_in
@@ -54,6 +64,14 @@ class ClassDef(object):
                 amf3 = True
         self.amf3 = amf3
 
+        if encode_types is None:
+            encode_types = {}
+        self.encode_types = encode_types
+
+        if decode_types is None:
+            decode_types = {}
+        self.decode_types = decode_types
+
     def getStaticAttrVals(self, obj):
         """Returns a list of values of attributes defined in self.static_attrs
 
@@ -87,8 +105,12 @@ class DynamicClassDef(ClassDef):
 
     DYNAMIC_CLASS_DEF = True
 
-    def __init__(self, class_, alias=None, static_attrs=None, amf3=True, _built_in=False):
-        ClassDef.__init__(self, class_, alias, static_attrs, amf3, _built_in)
+    def __init__(self, class_, alias=None, static_attrs=None, amf3=True,
+        encode_types=None, decode_types=None, include_private=None, _built_in=False):
+        ClassDef.__init__(self, class_, alias, static_attrs, amf3,
+            encode_types, decode_types, _built_in)
+
+        self.include_private = include_private
 
     def getDynamicAttrVals(self, obj, include_private=False):
         """Returns a dict where keys are attribute names and values are attribute values.
@@ -99,40 +121,48 @@ class DynamicClassDef(ClassDef):
         include_private - bool, if False do not include attributes with
             names starting with '_'. Default = False.
         """
-        return get_dynamic_attr_vals(obj, self.static_attrs, include_private=False);
+        if self.include_private is None:
+            ip = include_private
+        else:
+            ip = self.include_private
 
-class ExternizeableClassDef(ClassDef):
-    """A ClassDef where the byte string encoding/decoding is customized."""
+        return get_dynamic_attr_vals(obj, self.static_attrs, ip);
+
+class ExternClassDef(ClassDef):
+    """A ClassDef where the byte string encoding/decoding is customized.
+   
+    The Actionscript version of the class must implement IExternalizeable.
+    """
     
-    EXTERNIZEABLE_CLASS_DEF = True
+    EXTERNALIZABLE_CLASS_DEF = True
     
     def __init__(self, class_, alias=None, static_attrs=None, _built_in=False):
-        ClassDef.__init__(self, class_, alias, static_attrs, amf3=True, _built_in=_built_in)
+        ClassDef.__init__(self, class_, alias, static_attrs, amf3=True,
+            _built_in=_built_in)
  
-    def writeByteString(self, obj):
-        """Returns a byte string representation of the object.
-
+    def writeExternal(self, obj, context):
+        """
         This method must be overridden in a sub-class.
 
-        The return value can be a string or a ByteArray.
+        arguments
+        ==========
+         * obj - object, The object that is being encoded.
+         * context - amfast.decoder.EncoderContext, holds encoding related properties.
         """
         raise ClassDefError("This method must be implemented by a sub-class.")
 
-    def readByteString(self, obj, buf):
-        """Returns an integer specifying the
-        0-indexed position of the last decoded
-        byte in the byte array.
-
+    def readExternal(self, obj, context):
+        """
         This method must be overridden in a sub-class.
 
         arguments
         ==========
          * obj - object, The object that the byte string is being applied to.
-         * buf - string in 2.4 and 2.5, ByteArray in 2.6+, The bytes to be read.
+         * context - amfast.decoder.DecoderContext, holds decoding related properties.
         """
         raise ClassDefError("This method must be implemented by a sub-class.")
 
-class _ProxyClassDef(ExternizeableClassDef):
+class _ProxyClassDef(ExternClassDef):
     """A special class used internally to encode/decode Proxied objects."""
 
     PROXY_CLASS_DEF = True
@@ -143,7 +173,7 @@ class _ProxyClassDef(ExternizeableClassDef):
         pass
 
     def __init__(self):
-        ExternizeableClassDef.__init__(self, self._ProxyObject, self.PROXY_ALIAS,
+        ExternClassDef.__init__(self, self._ProxyObject, self.PROXY_ALIAS,
             None, _built_in=True)
 
 class _ArrayCollectionClassDef(_ProxyClassDef):
@@ -166,7 +196,7 @@ class _ObjectProxyClassDef(_ProxyClassDef):
 
 class ClassDefMapper(object):
     """Map classes to ClassDefs, retrieve class_defs by class or alias name."""
-    def __init__(self, class_def_attr='_amf_alias'):
+    def __init__(self):
         """
         arguments
         ==========
@@ -174,12 +204,15 @@ class ClassDefMapper(object):
             mapped classes. Default = '_amf_alias'
         """
         self._mapped_classes = {}
-        self.class_def_attr = class_def_attr
+        self._mapped_aliases = {}
         self._mapBuiltIns()
 
+    def __iter__(self):
+        return self._mapped_aliases.itervalues()
+
     def _mapBuiltIns(self):
-        """Map built-in ClassDefs."""
-        from amfast import remoting
+        """Map built-in ClassDefs for default behavior."""
+        from as_types import AsError
         from amfast.remoting import flex_messages as messaging
 
         # Proxy objects
@@ -187,13 +220,16 @@ class ClassDefMapper(object):
         self.mapClass(_ObjectProxyClassDef())
 
         # Exceptions
-        self.mapClass(ClassDef(remoting.AsError, _built_in=True))
+        self.mapClass(ClassDef(AsError, _built_in=True))
         self.mapClass(ClassDef(messaging.FaultError, _built_in=True))
 
         # Flex remoting messages
-        self.mapClass(ClassDef(messaging.AbstractMessage, _built_in=True))
         self.mapClass(ClassDef(messaging.RemotingMessage, _built_in=True))
+        self.mapClass(messaging.AsyncSmallMsgDef(messaging.AsyncMessage,
+            alias="DSA", _built_in=True))
         self.mapClass(ClassDef(messaging.AsyncMessage, _built_in=True))
+        self.mapClass(messaging.CommandSmallMsgDef(messaging.CommandMessage,
+            alias="DSC", _built_in=True))
         self.mapClass(ClassDef(messaging.CommandMessage, _built_in=True))
         self.mapClass(ClassDef(messaging.AcknowledgeMessage, _built_in=True))
         self.mapClass(ClassDef(messaging.ErrorMessage, _built_in=True))
@@ -208,8 +244,13 @@ class ClassDefMapper(object):
         if not hasattr(class_def, 'CLASS_DEF'):
             raise ClassDefError("class_def argument must be a ClassDef object.")
 
-        setattr(class_def.class_, self.class_def_attr, class_def)
-        self._mapped_classes[class_def.alias] = class_def
+        lock = threading.RLock()
+        lock.acquire()
+        try:
+            self._mapped_classes[class_def.class_] = class_def
+            self._mapped_aliases[class_def.alias] = class_def
+        finally:
+            lock.release()
 
     def getClassDefByClass(self, class_):
         """Get a ClassDef.
@@ -220,7 +261,14 @@ class ClassDefMapper(object):
         ==========
          * class_ - class, the class to find a ClassDef for.
         """
-        return getattr(class_, self.class_def_attr, None)
+        lock = threading.RLock()
+        lock.acquire()
+        try:
+            result = self._mapped_classes.get(class_, None)
+        finally:
+            lock.release()
+
+        return result
 
     def getClassDefByAlias(self, alias):
         """Get a ClassDef.
@@ -231,7 +279,14 @@ class ClassDefMapper(object):
         ==========
          * alias - string, the alias to find a ClassDef for.
         """
-        return self._mapped_classes.get(alias, None)
+        lock = threading.RLock()
+        lock.acquire()
+        try:
+            result = self._mapped_aliases.get(alias, None)
+        finally:
+            lock.release()
+
+        return result
 
     def unmapClass(self, class_):
         """Unmap a class definition.
@@ -240,22 +295,21 @@ class ClassDefMapper(object):
         ==========
          * class_ - class, the class to remove a ClassDef for.
         """
-        class_def = self.getClassDefByClass(class_)
-        if class_def is None:
-            raise ClassDefError("ClassDef not found.")
+        lock = threading.RLock()
+        lock.acquire()
+        try:
+            for alias, klass in self._mapped_aliases.iteritems():
+                if class_ == klass:
+                    del self._mapped_aliases[alias]
 
-        class_def = getattr(class_, self.class_def_attr)
-        delattr(class_, self.class_def_attr)
-        del self._mapped_classes[class_def.alias]
+            class_id = id(class_)
+            if class_id in self._mapped_classes:
+                del self._mapped_classes[class_id]
+        finally:
+            lock.release()
 
 # ---- module attributes ---- #
 
-# These properties can be set on a class
-# to automatically map attributes
-ALIAS = '_AMFAST_ALIAS'
-STATIC_ATTRS = '_AMFAST_STATIC_ATTRS'
-AMF3 = '_AMFAST_AMF3'
-
 def get_dynamic_attr_vals(obj, ignore_attrs=None, include_private=False):
     """Returns a dict of attribute values to encode.
 
@@ -282,6 +336,12 @@ def get_dynamic_attr_vals(obj, ignore_attrs=None, include_private=False):
 
     return vals
 
+# These properties can be set on a class
+# to map attributes within the class.
+ALIAS = '_AMFAST_ALIAS'
+STATIC_ATTRS = '_AMFAST_STATIC_ATTRS'
+AMF3 = '_AMFAST_AMF3'
+
 def assign_attrs(class_, alias=None, static_attrs=None, amf3=None):
     """
     Use to map ClassDef attributes to a class. Useful if you want to keep
