@@ -1,35 +1,18 @@
 """Provides an interface for performing remoting calls."""
+
+import threading
+
 import amfast
-from amfast import AmFastError, class_def, decoder, encoder
+from amfast import AmFastError, class_def
+from amfast.class_def.as_types import AsError
 
 class RemotingError(AmFastError):
     """Remoting related errors."""
     pass
 
-class AsError(RemotingError):
-    """Error object that is returned to the client.
-
-    Equivalent to: 'Error' in AS3.
-    """
-
-    APPLICATION_ERROR = 5000
-
-    def __init__(self, message='', exc=None):
-        self.errorID = self.APPLICATION_ERROR
-        if exc is not None:
-            self.name = exc.__class__.__name__
-            self.message = "%s" % exc
-        else:
-            self.name = ''
-            self.message = message
-
-        RemotingError.__init__(self, self.message)
-
-class_def.assign_attrs(AsError, 'Error', ('errorId', 'name', 'message'), False)
-
 class Service(object):
     """A remoting service is a service that is exposed 
-    by a Gateway to AMF clients. 
+    by an amfast.remoting.channel.Channel to AMF clients. 
 
     attributes:
     ============
@@ -39,9 +22,6 @@ class Service(object):
     # Name of special service that handles packet header targets
     PACKET_HEADER_SERVICE = 'PACKET_HEADER_SERVICE'
 
-    # Name of special service that handles message header targets
-    MESSAGE_HEADER_SERVICE = 'MESSAGE_HEADER_SERVICE'
-
     # Name of special service that handles command messages
     COMMAND_SERVICE = 'COMMAND_SERVICE'
 
@@ -52,45 +32,62 @@ class Service(object):
 
     def __init__(self, name):
         self.name = name
+
+        self._lock = threading.RLock()
         self._targets = {} # Keeps track of targets internally
 
-    def setTarget(self, target):
+    def __iter__(self):
+        return self._targets.itervalues()
+
+    def mapTarget(self, target):
         """Add a target to the service."""
-        self._targets[target.name] = target
+        self._lock.acquire()
+        try:
+            self._targets[target.name] = target
+        finally:
+            self._lock.release()
+
+    def unMapTarget(self, target):
+        """Remove a target from the service."""
+        self._lock.acquire()
+        try:
+            if target.name in self._targets:
+                del self._targets[target.name]
+        finally:
+            self._lock.release()
 
     def getTarget(self, target_name):
         """Get a target from the service by name."""
-        return self._targets.get(target_name, None)
-
-    def removeTarget(self, target):
-        """Remove a target from the service."""
-        if target.name in self._targets:
-            del self._targets[target.name]
+        self._lock.acquire()
+        try:
+            target = self._targets.get(target_name, None)
+        finally:
+            self._lock.release()
+        return target
 
 class Target(object):
-    """A remoting target can be invoked by a message.
+    """A remoting target can be invoked by an RPC message received from a client.
 
     attributes:
     ============
      * name - string, name of the target.
+     * secure - boolean, True to require login.
     """
-    def __init__(self, name):
+    def __init__(self, name, secure=False):
         self.name = name
+        self.secure = secure
 
-    def _invoke_str(self, args):
+    def _invokeStr(self, args):
         return "<targetInvocation target=\"%s\">%s</targetInvocation>" % \
             (self.name, args)
 
-    def invoke(self, request_packet, response_packet, request_msg, response_msg, args):
+    def invoke(self, packet, msg, args):
         """Invoke a target.
 
         arguments
         ==========
-         * request_packet - Packet, Packet that is invoking the target.
-         * response_packet - Packet, Packet that is being returned to the client.
-         * request_msg - Message, the message that is invoking this target.
-             This value will be None if the target is being invoked by a package header.
-         * response_msg - Message, the message that is being returned to the client.
+         * packet - Packet, Packet that is invoking the target.
+         * msg - Message, the message that is invoking this target.
              This value will be None if the target is being invoked by a package header.
          * args - list, list of arguments to pass to the callable.
         """
@@ -102,16 +99,17 @@ class CallableTarget(Target):
     attributes:
     ============
      * name - string, name of the target.
+     * secure - boolean, True to require login.
      * callable - callable, a callable that can be invoked.
     """
-    def __init__(self, callable, name):
-        Target.__init__(self, name)
+    def __init__(self, callable, name, secure=False):
+        Target.__init__(self, name, secure)
         self.callable = callable
 
-    def invoke(self, request_packet, response_packet, request_msg, response_msg, args):
+    def invoke(self, packet, msg, args):
         """Calls self.callable and passes *args."""
         if amfast.log_debug:
-            amfast.logger.debug(self._invoke_str(args))
+            amfast.logger.debug(self._invokeStr(args))
         return self.callable(*args)
 
 class ExtCallableTarget(CallableTarget):
@@ -120,12 +118,13 @@ class ExtCallableTarget(CallableTarget):
     attributes:
     ============
      * name - string, name of the target.
+     * secure - boolean, True to require login.
      * callable - callable, a callable that can be invoked.
     """
-    def invoke(self, request_packet, response_packet, request_msg, response_msg, args):
+    def invoke(self, packet, msg, args):
         if amfast.log_debug:
-            amfast.logger.debug(self._invoke_str(args))
-        return self.callable(request_packet, response_packet, request_msg, response_msg, *args)
+            amfast.logger.debug(self._invokeStr(args))
+        return self.callable(packet, msg, *args)
 
 class Header(object):
     """A remoting message header.
@@ -144,11 +143,12 @@ class Header(object):
     def __str__(self):
         return "<header name=\"%s\" required=\"%s\">%s</header>" % (self.name, self.required, self.value)
 
-    def invoke(self, service_mapper, request_packet, response_packet):
+    def invoke(self, request):
         """Invoke an action on this header if one has been mapped."""
-        target = service_mapper.packet_header_service.getTarget(self.name)
+        target = request.channel.channel_set.service_mapper.\
+            packet_header_service.getTarget(self.name)
         if target is not None:
-            return target.invoke(request_packet, response_packet, None, None, (self.value,))
+            return target.invoke(request, None, (self.value,))
         return False
 
 class Message(object):
@@ -158,17 +158,17 @@ class Message(object):
     ============
      * target - Target, the target to be invoked.
      * response - string, message id.
-     * value - object, message value.
+     * body - object, message body.
     """
 
     SUCCESS_TARGET = '/onResult'
     FAILED_TARGET = '/onStatus'
     DEBUG_TARGET = '/onDebugEvents'
 
-    def __init__(self, target=None, response=None, value=None):
+    def __init__(self, target=None, response=None, body=None):
         self.target = target
         self.response = response
-        self.value = value
+        self.body = body
 
     def _isInvokable(self):
         """If True, the message's body can invoke itself."""
@@ -177,24 +177,24 @@ class Message(object):
         return True
     is_invokable = property(_isInvokable)
 
-    def invoke(self, service_mapper, request_packet, response_packet):
-        """Invoke an action on this message and return a response message."""
+    def invoke(self, request):
+        """Invoke an action on an RPC message and return a response message."""
         try:
-            response_msg = self.acknowledge()
+            self.response_msg = self.acknowledge(request)
             if self.is_invokable:
-                self.value[0].invoke(service_mapper, request_packet, response_packet, self, response_msg)
+                self.body[0].invoke(request, self)
             elif self.target is not None and self.target != '':
-                self._invoke(service_mapper, request_packet, response_packet, response_msg)
+                self._invoke(request)
             else:
                 raise RemotingError("Cannot invoke message: '%s'." % self)
         except Exception, exc:
             amfast.log_exc()
-            response_msg = self.fail(exc)
+            self.response_msg = self.fail(request, exc)
 
-        return response_msg
+        return self.response_msg
 
-    def _invoke(self, service_mapper, request_packet, response_packet, response_msg):
-        """Invoke an action on an AMF0 style message."""
+    def _invoke(self, request):
+        """Invoke an action on an AMF0 style RPC message."""
         qualified_name = self.target.split(Service.SEPARATOR)
         if len(qualified_name) < 2:
             target_name = self.target
@@ -203,44 +203,49 @@ class Message(object):
             target_name = qualified_name.pop()
             service_name = Service.SEPARATOR.join(qualified_name)
 
-        target = service_mapper.getTargetByName(service_name, target_name)
+        target = request.channel.channel_set.service_mapper.getTarget(service_name, target_name)
         if target is None:
             raise RemotingError("Target '%s' not found." % self.target)
 
-        response_msg.value = target.invoke(request_packet, response_packet, self, response_msg, self.value)
+        if target.secure is True:
+            # Make sure user is authenticated
+            if not hasattr(request, '_authenticated'):
+                raise RemotingError('Target requires authentication.');
 
-    def fail(self, exc):
+        self.response_msg.body = target.invoke(request, self, self.body)
+
+    def fail(self, request, exc):
         """Return an error response message."""
         response_target = self.response + self.FAILED_TARGET
         response_message = Message(target=response_target, response='')
         
         if self.is_invokable:
-            error_val = self.value[0].fail(exc)
+            error_val = self.body[0].fail(request, self, exc)
         else:
             error_val = AsError(exc=exc)
 
-        response_message.value = error_val
+        response_message.body = error_val
         return response_message
 
-    def acknowledge(self):
-        """Return a successful response message."""
+    def acknowledge(self, request):
+        """Return a successful response message to acknowledge an RPC message."""
         response_target = self.response + self.SUCCESS_TARGET
         response_message = Message(target=response_target, response='')
         
         if self.is_invokable:
-            response_message.value = self.value[0].acknowledge()
+            response_message.body = self.body[0].acknowledge(request, self)
 
         return response_message
 
     def __str__(self):
-        return "<message> <target>%s</target> <response>%s</response> <value>%s</value></message>" % (self.target, self.response, self.value)
+        return "<message> <target>%s</target> <response>%s</response> <body>%s</body></message>" % (self.target, self.response, self.body)
 
 class Packet(object):
     """An AMF NetConnection packet that can be passed from client->server or server->client.
 
     attributes:
     ============
-     * version - string, the type of client connected to the server.
+     * client_type - string, the type of client connected to the server.
      * headers - dict, keys = header names, values = Header objects.
      * messages - list, a list of messages that belong to the packet.
     """
@@ -249,10 +254,10 @@ class Packet(object):
     FLASH_COM = "FLASH_COM"
     FLASH_9 = "FLASH_9"
 
-    def __init__(self, version=None, headers=None, messages=None):
-        if version is None:
-            version = self.FLASH_8
-        self.version = version
+    def __init__(self, client_type=None, headers=None, messages=None):
+        if client_type is None:
+            client_type = self.FLASH_8
+        self.client_type = client_type 
 
         if headers is None:
             headers = []
@@ -263,11 +268,49 @@ class Packet(object):
         self.messages = messages
 
     def _getAmf3(self):
-        if self.version == self.FLASH_9:
+        if self.client_type == self.FLASH_9:
             return True
         return False
     is_amf3 = property(_getAmf3)
 
+    def invoke(self):
+        """Process an RPC packet and return a response packet."""
+        if amfast.log_debug:
+            amfast.logger.debug("<requestPacket>%s</requestPacket>" % self)
+
+        self.response = self.acknowledge()
+        try:
+            # Invoke any headers
+            for header in self.headers:
+                header.invoke(self)
+
+            # Invoke any messages
+            for message in self.messages:
+                self.response.messages.append(message.invoke(self))
+        except Exception, exc:
+            # Fail all messages
+            amfast.log_exc()
+            self.response = self.fail(exc)
+
+        if amfast.log_debug:
+            amfast.logger.debug("<responsePacket>%s</responsePacket>" % self.response)
+
+        return self.response
+
+    def fail(self, exc):
+        """Return a response Packet with all messages failed."""
+        response = self.acknowledge()
+
+        for message in self.messages:
+            response.messages.append(message.fail(self, exc))
+        return response
+
+    def acknowledge(self):
+        """Create a response to this packet."""
+        response = Packet()
+        response.client_type = self.client_type
+        return response
+
     def __str__(self):
         header_msg = "\n  ".join(["%s" % header for header in self.headers])
 
@@ -284,128 +327,20 @@ class Packet(object):
  </messages>
 
  <attributes>
-  <attr name="version">%s</attr>
+  <attr name="client_type">%s</attr>
  </attributes>
 </Packet>
-""" % (header_msg, message_msg, self.version)
-
-    def invoke(self, service_mapper):
-        """Process this packet and return a response packet."""
-        if amfast.log_debug:
-            amfast.logger.debug("<requestPacket>%s</requestPacket>" % self)
-
-        response_packet = self.acknowledge()
-        try:
-            # Invoke any headers
-            for header in self.headers:
-                header.invoke(service_mapper, self, response_packet)
-
-            # Invoke any messages
-            for message in self.messages:
-                response_packet.messages.append(message.invoke(service_mapper, self, response_packet))
-        except Exception, exc:
-            # Fail all messages
-            amfast.log_exc()
-            response_packet = self.fail(exc)
-
-        if (response_packet.messages is None or len(response_packet.messages) == 0) and \
-            (response_packet.headers is None or len(response_packet.headers) == 0):
-            # Empty response
-            response_packet = None
-
-        if amfast.log_debug:
-            amfast.logger.debug("<responsePacket>%s</responsePacket>" % response_packet)
-
-        return response_packet
-                
-    def fail(self, exc):
-        """Return a response Packet with all messages failed."""
-        response_packet = self.acknowledge()
-
-        for message in self.messages:
-            response_packet.messages.append(message.fail(exc))
-        return response_packet
-
-    def acknowledge(self):
-        """Create a response to this packet."""
-        response_packet = Packet()
-        response_packet.version = self.version
-        return response_packet
-
-class Gateway(object):
-    """An AMF remoting gateway."""
-    def __init__(self, service_mapper=None, class_def_mapper=None,
-        use_array_collections=False, use_object_proxies=False,
-        use_references=True, use_legacy_xml=False, include_private=False):
-
-        self.service_mapper = service_mapper
-        if self.service_mapper is None:
-            self.service_mapper = ServiceMapper()
-
-        self.class_def_mapper = class_def_mapper
-        if self.class_def_mapper is None:
-            self.class_def_mapper = class_def.ClassDefMapper()
-
-        self.use_array_collections = use_array_collections
-        self.use_object_proxies = use_object_proxies
-        self.use_references = use_references
-        self.use_legacy_xml = use_legacy_xml
-        self.include_private = include_private
-
-    def process_packet(self, raw_packet):
-        """Process an incoming packet."""
-        if amfast.log_debug:
-            amfast.logger.debug("<gateway>Processing incoming packet.</gateway>")
-
-        request_packet = None
-        try:
-            request_packet = self.decode_packet(raw_packet)
-            response_packet = request_packet.invoke(self.service_mapper)
-            if response_packet is None:
-                return None
-            else:
-                return self.encode_packet(response_packet, request_packet.is_amf3)
-        except Exception, exc:
-            amfast.log_exc()
-
-            if request_packet is not None:
-               return self.encode_packet(request_packet.fail(exc), request_packet.is_amf3)
-            else:
-                # There isn't much we can do if
-                # the request was not decoded correctly.
-                raise exc
-
-    def decode_packet(self, raw_packet):
-        if amfast.log_debug:
-            amfast.logger.debug("<rawRequestPacket>%s</rawRequestPacket>" %
-                amfast.format_byte_string(raw_packet))
-
-        return decoder.decode(raw_packet, packet=True,
-            class_def_mapper=self.class_def_mapper)
-
-    def encode_packet(self, packet, amf3=False):
-        raw_packet = encoder.encode(packet, packet=True, 
-            class_def_mapper=self.class_def_mapper,
-            use_array_collections=self.use_array_collections,
-            use_object_proxies=self.use_object_proxies,
-            use_references=self.use_references, use_legacy_xml=self.use_legacy_xml,
-            include_private=self.include_private, amf3=amf3)
-
-        if amfast.log_debug:
-            amfast.logger.debug("<rawResponsePacket>%s</rawResponsePacket>" %
-                amfast.format_byte_string(raw_packet))
-
-        return raw_packet
+""" % (header_msg, message_msg, self.client_type)
 
 class ServiceMapper(object):
     """Maps service to service name.
 
     attributes
     ===========
-    packet_header_service - Service, a special service for AMF packet headers.
-    message_header_service - Service, a special service for AMF message headers.
-    command_service - Service, a special service for Flex CommandMessages.
-    default_service - Service, a special service for targets that don't have service specifiers.
+     * packet_header_service - Service, a special service for AMF packet headers.
+     * message_header_service - Service, a special service for AMF message headers.
+     * command_service - Service, a special service for Flex CommandMessages.
+     * default_service - Service, a special service for targets that don't have service specifiers.
     
     When an AMF packet or message is processed, the object checks
     the header_service Service for a target where target.name == header.name
@@ -418,25 +353,47 @@ class ServiceMapper(object):
     """
 
     def __init__(self):
-        self._mapped_services = {} # used internally to keep track of Service objects.
+        self._services = {} # used internally to keep track of Service objects.
         self._mapBuiltIns()
 
+    def __iter__(self):
+        return self._services.itervalues()
+
     def _mapBuiltIns(self):
+        """Map default Targets required for Authentication and FlexMessaging.
+
+        Users can override the defaults, by remapping their own targets.
+        """ 
         import amfast.remoting.flex_messages as messaging
         import targets
 
         # Map built in targets
         self.packet_header_service = Service(Service.PACKET_HEADER_SERVICE)
         self.mapService(self.packet_header_service)
-        self.message_header_service = Service(Service.MESSAGE_HEADER_SERVICE)
-        self.mapService(self.message_header_service)
         self.command_service = Service(Service.COMMAND_SERVICE)
         self.mapService(self.command_service)
         self.default_service = Service(Service.DEFAULT_SERVICE)
         self.mapService(self.default_service)
 
-        self.command_service.setTarget(ExtCallableTarget(targets.ro_ping,
+        # NetConnection authentication
+        self.packet_header_service.mapTarget(ExtCallableTarget(targets.nc_auth,
+            'Credentials'))
+
+        # CommandMessages
+        self.command_service.mapTarget(ExtCallableTarget(targets.client_ping,
             messaging.CommandMessage.CLIENT_PING_OPERATION))
+        self.command_service.mapTarget(ExtCallableTarget(targets.login_operation,
+            messaging.CommandMessage.LOGIN_OPERATION))
+        self.command_service.mapTarget(ExtCallableTarget(targets.logout_operation,
+            messaging.CommandMessage.LOGOUT_OPERATION))
+        self.command_service.mapTarget(ExtCallableTarget(targets.poll_operation,
+            messaging.CommandMessage.POLL_OPERATION))
+        self.command_service.mapTarget(ExtCallableTarget(targets.subscribe_operation,
+            messaging.CommandMessage.SUBSCRIBE_OPERATION))
+        self.command_service.mapTarget(ExtCallableTarget(targets.unsubscribe_operation,
+            messaging.CommandMessage.UNSUBSCRIBE_OPERATION))
+        self.command_service.mapTarget(ExtCallableTarget(targets.disconnect_operation,
+            messaging.CommandMessage.DISCONNECT_OPERATION))
 
     def mapService(self, service):
         """Maps a service
@@ -445,7 +402,12 @@ class ServiceMapper(object):
         ==========
          * service - Service, the service to map.
         """
-        self._mapped_services[service.name] = service
+        lock = threading.RLock()
+        lock.acquire()
+        try:
+            self._services[service.name] = service
+        finally:
+            lock.release()
 
     def unMapService(self, service):
         """Un-maps a service
@@ -454,10 +416,15 @@ class ServiceMapper(object):
         ==========
          * service - Service, the service to un-map.
         """
-        if service.name in self._mapped_services:
-            del self._mapped_services[service.name]
+        lock = threading.RLock()
+        lock.acquire()
+        try:
+            if service.name in self._services:
+                del self._services[service.name]
+        finally:
+            lock.release()
 
-    def getTargetByName(self, service_name, target_name):
+    def getTarget(self, service_name, target_name):
         """Get a Target
 
         Returns None in Target is not found.
@@ -467,13 +434,20 @@ class ServiceMapper(object):
          * service_name - string, the service name.
          * target_name - string, the target name.
         """
-        service = self.getServiceByName(service_name)
-        if service is None:
-            return None
+        lock = threading.RLock()
+        lock.acquire()
+        try:
+            service = self.getService(service_name)
+            if service is None:
+                target = None
+            else:
+                target = service.getTarget(target_name)
+        finally:
+            lock.release()
 
-        return service.getTarget(target_name)
+        return target
 
-    def getServiceByName(self, service_name):
+    def getService(self, service_name):
         """Get a Service
 
         Returns None in Service is not found.
@@ -482,4 +456,11 @@ class ServiceMapper(object):
         ==========
          * service_name - string, the service name.
         """
-        return self._mapped_services.get(service_name, None)
+        lock = threading.RLock()
+        lock.acquire()
+        try:
+            service = self._services.get(service_name, None)
+        finally:
+            lock.release()
+
+        return service
